module jl
  use Precision
  implicit none

  private
  public integratejl

  contains


    SUBROUTINE integratejl(l,a,b,s)
      INTEGER JMAX
      REAL(dl) a,b,s,EPS
      PARAMETER (EPS=1.e-4, JMAX=20)
      INTEGER j,l 
      REAL(dl) os,ost,st
      ost=-1.e30
      os= -1.e30
      do j=1,JMAX
         call trapzd(a,b,st,j)
         s=(4.*st-ost)/3.
         if (abs(s-os).lt.EPS*abs(os)) return
         if (abs(s-os).lt.EPS) return
         os=s
         ost=st
      end do

      contains

        SUBROUTINE trapzd(a,b,s,n)
          INTEGER n
          REAL(dl) a,b,s
          INTEGER it,j
          REAL(dl) del,sum,tnm,x
          if (n.eq.1) then
             s=0.5*(b-a)*(func(a)+func(b))
          else
             it=2**(n-2)
             tnm=it
             del=(b-a)/tnm
             x=a+0.5*del
             sum=0.
             do j=1,it
                sum=sum+func(x)
                x=x+del
             enddo
             s=0.5*(s+(b-a)*sum/tnm)
          end if
          return
        END SUBROUTINE trapzd

        real(dl) function func(x)
          real(dl) :: x, jl
          call bjl(l, x, jl)

          func = jl
          return
        end function func

    END SUBROUTINE integratejl

! jl code from CAMB, based on the approximations by Kosowsky
    

    subroutine bjl(l,x,jl)
         use Precision
!  Calculates the spherical bessel function j_l(x)

!  and optionally its derivative for real x and integer l>=0.

!  Asymptotic approximations 8.11.5, 8.12.5, and 8.42.7 from

!  G.N.Watson, A Treatise on the Theory of Bessel Functions,
!  2nd Edition (Cambridge University Press, 1944).
!  Higher terms in expansion for x near l given by
!  Airey in Phil. Mag. 31, 520 (1916).

!  This approximation is accurate to near 0.1% at the boundaries
!  between the asymptotic regions; well away from the boundaries
!  the accuracy is better than 10^{-5}. The derivative accuracy
!  is somewhat worse than the function accuracy but still better
!  than 1%.

!  Point *jlp initially to a negative value to forego calculating
!  the derivative; point it to a positive value to do the derivative
!  also (Note: give it a definite value before the calculation
!  so it's not pointing at junk.) The derivative calculation requires

!  only arithmetic operations, plus evaluation of one sin() for the
!  x>>l region.


!  Original code by Arthur Kosowsky   akosowsky@cfa.harvard.edu
!  This fortran version only computes j_l(x)
        implicit real(dl)(a-h,o-z)
        integer l
        real(dl) nu, nu2,ax,ax2,beta,beta2,beta4,beta6
        real(dl) sx,sx2,cx,sum1,sum2,sum3,sum4,sum5,deriv1
        real(dl) cotb,cot3b,cot6b,secb,sec2b,sec4b,sec6b
        real(dl) trigarg,trigcos,expterm,prefactor,llimit,ulimit,fl
        real(dl) x,jl

!        PI = 3.1415926536
        ROOTPI = 1.772453851
        GAMMA1 = 2.6789385347             !/* Gamma function of 1/3 */
        GAMMA2 = 1.3541179394             !/* Gamma function of 2/3 */


        ax = abs(x)
        fl = l


        beta = fl**0.325
        llimit=1.31*beta   !/* limits of asymptotic regions; fitted */
        ulimit=1.48*beta

         nu= fl + 0.5

         nu2=nu*nu

        if (l .lt. 0) then
                print*, 'Bessel function index < 0\n'
                stop
        end if

!          /************* Use closed form for l<6 **********/

        if (l .lt. 6) then

        sx=sin(ax)
        cx=cos(ax)
        ax2=ax*ax

            if(l .eq. 0) then
            if(ax .gt. 0.001) then
                 jl=sx/ax
                 else

                 jl=1._dl-ax2/6._dl
                 end if    !   /* small x trap */
            endif


            if(l .eq. 1) then

            if(ax .gt. 0.001) then
                 jl=(sx/ax -cx)/ax
                 else

                 jl=ax/3._dl
                 end if
            endif

            if(l .eq. 2) then
              if(ax .gt. 0.001) then
                  jl=(-3._dl*cx/ax &
                          -sx*(1._dl-3._dl/ax2))/ax
              else

                  jl=ax2/15._dl
                  end if
            endif

            if(l .eq. 3) then
          if(ax .gt. 0.001) then
                jl=(cx*(1._dl-15._dl/ax2) &
                          -sx*(6._dl-15._dl/ax2)/ax)/ax
           else

                jl=ax*ax2/105._dl
                endif
            endif

            if(l .eq. 4) then
          if(ax .gt. 0.001) then
        jl=(sx*(1._dl-45._dl/(ax*ax)+105._dl &
             /(ax*ax*ax*ax)) +cx*(10._dl-105._dl/(ax*ax))/ax)/ax
          else

                jl=ax2*ax2/945._dl
                end if
            endif

             if(l .eq. 5) then

          if(ax .gt. 0.001) then
        jl=(sx*(15._dl-420._dl/(ax*ax)+945._dl &
          /(ax*ax*ax*ax))/ax -cx*(1.0-105._dl/(ax*ax)+945._dl &
                                          /(ax*ax*ax*ax)))/ax
           else

                jl=ax2*ax2*ax/10395._dl
                endif
             endif


!          /********************** x=0 **********************/

        else if (ax .lt. 1.d-30) then
        jl=0.0

!          /*************** Region 1: x << l ****************/

        else if (ax .le. fl+0.5-llimit) then


!       beta=acosh(nu/ax)
        if (nu/ax .lt. 1._dl) print*, 'trouble with acosh'
        beta = dlog(nu/ax + sqrt((nu/ax)**2 - 1._dl) )
                !(4.6.21)
        cotb=nu/sqrt(nu*nu-ax*ax)      ! /* cotb=coth(beta) */
        cot3b=cotb*cotb*cotb
        cot6b=cot3b*cot3b
        secb=ax/nu
        sec2b=secb*secb
        sec4b=sec2b*sec2b
        sec6b=sec4b*sec2b
        sum1=2.0+3.0*sec2b
        expterm=sum1*cot3b/(24.0*nu)
        sum2=4.0+sec2b
        expterm = expterm - sum2*sec2b*cot6b/(16.0*nu2)
        sum3=16.0-1512.0*sec2b-3654.0*sec4b-375.0*sec6b
        expterm = expterm - sum3*cot3b*cot6b/(5760.0*nu*nu2)
        sum4=32.0+288.0*sec2b+232.0*sec4b+13.0*sec6b
        expterm = expterm - sum4*sec2b*cot6b*cot6b/(128.0*nu2*nu2)
        expterm=exp(-nu*beta+nu/cotb-expterm)
        prefactor=sqrt(cotb/secb)/(2.0*nu)
        jl=prefactor*expterm

!          /**************** Region 2: x >> l ****************/


        else if (ax .ge. fl+0.5+ulimit) then


        beta=acos(nu/ax)
        cotb=nu/sqrt(ax*ax-nu*nu)      !/* cotb=cot(beta) */
        cot3b=cotb*cotb*cotb
        cot6b=cot3b*cot3b
        secb=ax/nu
        sec2b=secb*secb
        sec4b=sec2b*sec2b
        sec6b=sec4b*sec2b
        trigarg=nu/cotb - nu*beta - PI/4.0
        sum1=2.0+3.0*sec2b
        trigarg = trigarg - sum1*cot3b/(24.0*nu)
        sum3=16.0-1512.0*sec2b-3654.0*sec4b-375.0*sec6b
        trigarg = trigarg - sum3*cot3b*cot6b/(5760.0*nu*nu2)
        trigcos=cos(trigarg)
        sum2=4.0+sec2b
        expterm=sum2*sec2b*cot6b/(16.0*nu2)
        sum4=32.0+288.0*sec2b+232.0*sec4b+13.0*sec6b
        expterm = expterm - sum4*sec2b*cot6b*cot6b/(128.0*nu2*nu2)
        expterm=exp(-expterm)
        prefactor=sqrt(cotb/secb)/nu
        jl=prefactor*expterm*trigcos

!          /***************** Region 3: x near l ****************/

        else



        beta=ax-nu

        beta2=beta*beta
        beta4=beta2*beta2
        beta6=beta2*beta4
        sx=6.0/ax
        sx2=sx*sx
        cx=sqrt(sx)

        secb=sx**0.333333333

        sec2b=secb*secb

        deriv1=GAMMA1*secb
        deriv1= deriv1+ beta*GAMMA2*sec2b
        sum1=(beta2/6.0-1.0/15.0)*beta
        deriv1 = deriv1 - sum1*sx*secb*GAMMA1/3.0
        sum2=beta4/24.0-beta2/24.0+1.0/280.0
        deriv1 = deriv1 - 2.0*sum2*sx*sec2b*GAMMA2/3.0
        sum3=beta6/720.0-7.0*beta4/1440.0+beta2/288.0-1.0/3600.0
        deriv1 = deriv1 + 4.0*sum3*sx2*secb*GAMMA1/9.0
        sum4=(beta6/5040.0-beta4/900.0+19.0*beta2/12600.0- &
        13.0/31500.0)*beta
        deriv1 = deriv1 + 10.0*sum4*sx2*sec2b*GAMMA2/9.0
        sum5=(beta4*beta4/362880.0-beta6/30240.0+71.0*beta4/604800.0 &
                     -121.0*beta2/907200.0 + 7939.0/232848000.0)*beta
        deriv1 = deriv1 - 28.0*sum5*sx2*sx*secb*GAMMA1/27.0

        jl=deriv1*cx/(12.0*ROOTPI)

        end if

      end subroutine bjl



end module jl
